#include "COMP1.h"

uint8_t _initialized;
uint8_t _interruptEnabled;
      
//ISR (Interrupt Service Routine) called by the analog comparator when
//  the user choose the raise of an interrupt
//void cmp1_isr() { // nur zum Test, ob Interrupts bearbeitet werden
//    if (CMP1_SCR & CMP_SCR_CFR) digitalWrite(7,HIGH);
//    if (CMP1_SCR & CMP_SCR_CFF) digitalWrite(7,LOW);
//    CMP1_SCR |= CMP_SCR_CFR | CMP_SCR_CFF; // clear CFR and CFF
//}

void CMP1_Init(void) {
    NVIC_SET_PRIORITY(IRQ_CMP1, 130); // 0 = highest priority, 255 = lowest
    NVIC_ENABLE_IRQ(IRQ_CMP1); // handler is now cmp1_isr()
    // CMP1_IN0 (pin23), sicherheitshalber:
    CORE_PIN23_CONFIG = PORT_PCR_MUX(0); // set pin23 function to ALT0
}

//setting and switching on the analog comparator
uint8_t Comp1SetOn(uint8_t Psel, uint8_t Nsel) {
    if (_initialized) { //already running
        return 1;
    }

    // clock gate: comparator clock set on. it is off by default for conserving power.
    SIM_SCGC4 |= SIM_SCGC4_CMP;
    
    CMP1_CR1 = 0; // set CMPx_CR1 to a known state
    CMP1_CR0 = 0; // FILTER_CNT=HYSTCTR=0
    CMP1_CR1 = B00010001; // SE=WE=0, PMODE=1, INV=COS=OPE=0, EN=1
    CMP1_SCR = B00000110; // DMAEN=IER=IEF=0, clear flags CFR,CFF
    CMP1_FPR = 255; // max.Filterdelay
    CMP1_MUXCR = Psel << 3 | Nsel; // set comparator inputs
    CMP1_DACCR = B11010100; // B11011110: enable, Vrsel=1, val=30 (1.6 V)

    _initialized = 1;
    return 0; //OK
}

//switch off the analog comparator
void Comp1SetOff() {
    if (_initialized) {
        if (_interruptEnabled) {
            CMP1_SCR = 0;         //disable interrupts on AC events
            _interruptEnabled = 0;
        }
        CMP1_CR1 = 0; //switch off the AC
        _initialized = 0;
    }
}

//enable the interrupt on comparations
void Comp1EnableIntrpt(uint8_t tempMode) {
    if (_interruptEnabled) { //disable interrupts
        CMP1_SCR = 0;
    }

    if (!_initialized) {
        Comp1SetOn(0, 7);
    }

    //set the interrupt mode
    if (tempMode == CHANGE) {
        CMP1_SCR = CMP_SCR_IER | CMP_SCR_IEF;
    } else if (tempMode == FALLING) {
        CMP1_SCR = CMP_SCR_IEF;
    } else { //default is RISING
        CMP1_SCR = CMP_SCR_IER;
    }

    _interruptEnabled = 1;
}


//disable the interrupt on comparations
void Comp1DisableIntrpt(void) {
    if ((!_initialized) || (!_interruptEnabled)) {
        return;
    }
    CMP1_SCR = 0; //disable interrupt
    _interruptEnabled = 0;
}
