#ifndef _myOregonDecoderV3_2_
#define _myOregonDecoderV3_2_

#include <Arduino.h>
#include <digitalWriteFast.h>

#define WGR800    // auskommentiert heißt TFA-Sensor

#ifdef WGR800
#define PW1 350     // Original 200
#define PW2 720     // Original 700
#define PW3 1150     // Original 1200
#define FLIP 32
#define TOTALB 80
#define N_DATASETS 4  // 1 mehr als TFA für Temperatursensor
#else       // TFA
#define PW1 200
#define PW2 750
#define PW3 1200
#define FLIP 5
#define TOTALB 104
#define N_DATASETS 6  // TFA-Windsensor sendet Daten 3 mal
#endif
#define L_DATASET 15  // 15 reicht für TFA und WGR800

// 433 MHz Oregon-decoder:

class OOK_OregonDecodeV3_2 {
  // Die Klassen DecodeOOK und OregonDecoderV3 sind hier zusammengefasst

protected:
    byte bits;
    volatile byte data[N_DATASETS][L_DATASET];
    volatile byte* data_pnt; // wird auf eine der N_DATASETS Datenbänke zeigen
    char dc;
public:
    volatile byte flip; // alles hier wegen Direktzugriff ...
    volatile byte state;
    volatile byte total_bits;
    volatile byte pos;  // ... hier wegen Direktzugriff

    word anz_p=0;   // zählt die Pulse pro Datensatz
    enum { UNKNOWN, T0, OK, DONE };

    OOK_OregonDecodeV3_2 () {
      resetDecoder(); 
    }

    bool isDone () {
      return state == DONE;
    }
   
    // Pointer auf einen der Datenblöcke setzen:
    void setdb(byte rpos) {
      data_pnt = data[rpos];
    }
    
    volatile const byte* getData (byte rpos) const {
        return data[rpos]; 
    }
   
    void resetDecoder () {
        bits = flip = 0;
        total_bits = pos = 0;
        anz_p = 0;
        state = UNKNOWN;
        digitalWrite(A2,LOW);  // flip>0
        digitalWrite(A3,LOW);  // total_bits>0
    }
    
    // add one bit to the packet data buffer (Methode V3)
    void gotBit (char value) {
        if (pos >= L_DATASET) {  // im Original war diese Abfrage am Ende,
            resetDecoder();       // aber hier sicher vor Ueberschreiben geschuetzt
            return;
        }
        data_pnt[pos] = (data_pnt[pos] >> 1) | (value ? 0x80 : 00); // LSB comes first
        total_bits++;
        pos = total_bits >> 3;
        if (++bits >= 8)
            bits = 0;
        state = OK;
        digitalWrite(A3,HIGH);  // total_bits>0, start with data bits
    }
    
    // store a bit using Manchester encoding
    void manchester (char value) {
        flip ^= value; // manchester code, long pulse flips the bit
        gotBit(flip);
    }

    void done () {
        while (bits)
            gotBit(0); // padding
        state = DONE;
    }
    
    bool nextPulse (word width) {
        if (state != DONE) {
            anz_p++;
//            switch (decode(width)) {
//                case -1: resetDecoder();  break;  // OregonDecodeV3_2.h:116:17: warning: 
//                                                  // case label value is less than minimum 
//                                                  // value for type ??
//                case 1:  done(); break;
//            }
            dc=decode(width);
            if (dc==-1) resetDecoder();
            else if (dc==1) done();
        }
        return isDone();
    }

    char decode (word width) {
        if (PW1 < width && width < PW3) {
            byte w = width >= PW2;
            switch (state) {
                case UNKNOWN:
                    if (w == 0) {
                        ++flip;
                        digitalWrite(A2,HIGH);  // flip>0
                    }
                    else if (FLIP <= flip) {
                        flip = 1;
                        manchester(1);
                    } else
                        return -1;
                    break;
                case OK:
                    if (w == 0)
                        state = T0;
                    else
                        manchester(1);
                    break;
                case T0:
                    if (w == 0)
                        manchester(0);
                    else
                        return -1;
                    break;
            }
        } else 
            return -1;

        return total_bits == TOTALB ? 1: 0;
    }

};

#endif
