/* Test der Decodierung der Oregon-Daten (WindEmpfSend_t32) 
   auf einem Teensy 3.2
   */


// Decodierung Oregon-Daten:
#include <Arduino.h>
#include <stdio.h>
#include <math.h>
#include <digitalWriteFast.h>
#include "OregonDecodeV3_2.h"
#include "COMP1.h"

#include <util/atomic.h>

#define DEBUG_LED       13
#define SMS_LEN         160

// Decodierung Oregon-Daten:
#define INTROUTPIN A5   // zur Darstellung der Intr.-Pulse
#define MARKSENT   A1  // zeigt 1/2 Sek. lang an LED an, dass Datenblock gesendet wurde

OOK_OregonDecodeV3_2 orscV3;
boolean SerMonitor;

// Decodierung Oregon-Daten:
bool tglbit=false;
unsigned long tm, talkInterv; // Zeit für einen loop-Durchlauf
// Winddata:
//word wv;
byte lowbatt;
// for temperature data:
int temp_act;
byte rh_act;

// Parameter in loop():
byte n, dtyp;
uint16_t Scount=0;
short int v16=30;  // muss static sein, damit Intr Datenblock weiterschaltet!?
byte v8=6;         // muss static sein, damit Intr Datenblock weiterschaltet!?
char outstr[12];
unsigned long tm_wd, tloop;  // Zeit zwischen WGR800-Winddaten
unsigned long ts;
byte norecdat;        // Luecke zwischen Winddaten in 14sec-Einheiten
char ic='n';
byte respos_cpy, respos_cold=3;
byte totp_c;
byte dbcnt_res;
byte intrPin;        // Interrupt-Eingang
uint16_t cnt_tot=0, cnt_unk=0;

// Parameter, die von der Komparator-Interrupt-Routine veraendert werden koennen:
volatile byte res_valid[N_DATASETS];
volatile byte dbcount[N_DATASETS];
volatile byte res_pos=0;
volatile uint16_t dtmin=255, dtmax=0;
volatile word pulse, tot_p;

volatile uint16_t pulswidths[N_DATASETS][200];
volatile byte idx[N_DATASETS];
volatile byte pwidx=0;
volatile byte old_state=0;

void waitSM(const char* msg) {
  Serial.println(msg);
  while (Serial.available()>0) Serial.read();
  while (Serial.available()==0);
  while (Serial.available()>0) Serial.read();
}

// Interrupt Routine
/* The output of the 433 MHz receiver is connected to the input of comparator 1.
   The interrupt routine measures the pulse widths, stores them into a ring buffer,
   decodes the Manchester coded signal, and stores the decoded bits into a data buffer.
   There are N_DATASETS=4 (defined in OregonDeocdeV3_2.h) data buffers and ring buffers.
   They are selected by the variable res_pos which switches to the next buffers every
   time when a data set is complete.
   A data set consists of 80 bits which are received within about 0.1 sec. The time gap
   between data sets is about 7 sec.  
   Of course the 433 MHz receiver will not only see the data signal but also many other
   RF signals on the air.  */
void cmp1_isr() {   // 433 MHz-Empfaenger ist an CMP1_IN0, pin 23, angeschlossen
    static unsigned long last, pw, t00; // mit static muss Speicher für die Parameter
                                       //   nur einmal reserviert werden
    static uint16_t dt;
    
    if (CMP1_SCR & CMP_SCR_CFR) digitalWrite(INTROUTPIN,HIGH); // check by oscilloscope
    if (CMP1_SCR & CMP_SCR_CFF) digitalWrite(INTROUTPIN,LOW);
    CMP1_SCR |= CMP_SCR_CFR | CMP_SCR_CFF; // clear CFR and CFF
    
    // determine the pulse length in microseconds, for either polarity
    t00 = micros();   // nur zum Testen der Intr.-Dauer
    pw = micros() - last;
    last += pw;
    if (pw < 65536) pulse=(word)pw;
    else pulse=65535;

      // store pulses into ring buffer:
//      pulswidths[res_pos][pwidx++]=pulse; // stored and printed later for simulation tests
//      if (pwidx > 199) pwidx=0;

    // decode signal
//    atomic_thread_fence(std::memory_order_acquire ) {
      if (orscV3.nextPulse(pulse)==1) {
        dbcount[res_pos] = orscV3.pos;  // pos retten, ehe es zu 0 wird
        res_valid[res_pos]=1;  // valid Flag
        idx[res_pos]=pwidx; // save position of the last pulse
        if (orscV3.state != old_state) {
          digitalWrite(6+old_state,LOW);
          digitalWrite(6+orscV3.state,HIGH); // das ist DONE
          old_state=orscV3.state;
        }
        orscV3.resetDecoder ();
        
        pwidx=0;  // müsste nicht sein
        digitalWrite(2+res_pos,LOW);
        res_pos++;
        if (res_pos > (N_DATASETS-1)) res_pos = 0;
        digitalWrite(2+res_pos,HIGH);
        orscV3.setdb(res_pos);  // switch to next buffers
//        if (res_pos == 1) { ich versteh' die Ergebnisse daraus nicht
//          t00 = millis();
//        }
//        else {
//          dtmax=millis()-t00;
//        }
      }
//      else {  // cures compiler/optimizer failure
//        if (orscV3.flip>1 && pulse>PW2) orscV3.resetDecoder();
//      }
      if (orscV3.state != old_state) {
        digitalWrite(6+old_state,LOW);
        digitalWrite(6+orscV3.state,HIGH);
        old_state=orscV3.state;
      }
//    }
    dt=micros()-t00;
    if (dt<dtmin) dtmin=dt;
    if (dt>dtmax) dtmax=dt;
}

byte readRFData(byte r_pos, short int* val16, byte* val8, byte* lowBatt, boolean mean_speed, boolean prt2) {
  // liest Daten des Empfaengers aus dem Datensatz r_pos und liefert sie in val16 und val8 zurück;
  //  mean_speed==true: gemittelte Windgeschwindigkeit
  //  mean_speed==false: aktuelle Windgeschwindigkeit
  //            WGR800        THGN          TFA    
  // val16    Geschw.*10    Temperatur*10  Geschw.*10 simuliert
  // val8     Richtung      Luftfeuchte    Richtung simuliert
  // return     1             0             2
  // 
  volatile const byte* data;
  short int  wv, tmp;   // für Geschw.,  Temp. 
  byte wd, rh;              // für Richtung, rel. Luftfeuchte
  byte retv=0;

//Comp1DisableIntrpt(); // cli();
data = orscV3.getData(r_pos); // this should work without cli/sei because data are accessed
                              // only after a completed data set when the interrupt routine
                              // is using one of the next buffers.
                              // The 4 buffers are covering a time period of 4*7 sec. The
                              // main loop takes about 1 sec.
//Comp1EnableIntrpt(CHANGE);  // sei();
//Serial.print("von Datensatz ");Serial.print(r_pos);Serial.print(" ");
  if ( (data[0] == 26) & (data[1] == 137) ) { // WGR800: 1A89
    if (mean_speed)
      wv = (data[8] >> 4)*100 + (data[8] & 0x0F)*10 + (data[7] >> 4);
    else
      wv = (data[7] & 0x0F)*100 + (data[6] >> 4)*10 + (data[6] & 0x0F);
//    Serial.print(": ");Serial.print(wv);Serial.print("m/s*10 ");
//    wv = word(round(wv * 3.6));   // kmh*10
    wd = (data[4] >> 4);

    *val16 = wv;
    *val8 = wd;

    if (data[4] & 0x04)
      *lowBatt = 4;   // battery low Windsensor
    else
      *lowBatt = 0;
      
    retv = 1;
  }
  else if ( (data[0] == 250) && (data[1] == 40) ) {  // THGR810: FA28
    tmp = (data[5] >> 4)*100 + (data[5] & 0x0F)*10 + (data[4] >> 4);
    if (data[6] & 0x0F) tmp = -tmp;
    rh = (data[7] & 0x0F)*10 + (data[6] >> 4);

    *val16 = tmp;
    *val8 = rh;
    
    if (data[4] & 0x04)
      *lowBatt = 5;   // battery low Temperatursensor
    else
      *lowBatt = 0;
      
    retv = 0;
  }
  else {  // unbekannte Daten
    retv = 2;

    if (prt2) {
      for (byte i=0; i<10; i++)
        Serial.print(data[i],HEX);
    }
  }
    
  return(retv);
}

void mark_dataReceived() {
  digitalWrite(MARKSENT,HIGH);
  delay(500);
  digitalWrite(MARKSENT,LOW);
  delay(200);
}


void setup() { // -------------------------------------------------------------
  
  Serial.begin(38400);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }
  Serial.println(F("<<< Oregon-Decodierung >>>"));

  SerMonitor=true;

  pinMode(INTROUTPIN,OUTPUT);  // fuer Anzeige in Interrupt-Routine
  pinMode(MARKSENT,OUTPUT);
  pinMode(A0,OUTPUT); // fuer beep
  digitalWrite(MARKSENT,LOW); // LED aus

  // Pins fuer Echtzeitanalyse:
  pinMode(A2,OUTPUT);  // flip>0
  pinMode(A3,OUTPUT);  // total_bits>0
  pinMode(A4,OUTPUT);  // case2
  // A5 ist  INTROUTPIN
  pinMode(2,OUTPUT);  // res_pos0
  pinMode(3,OUTPUT);  // res_pos1
  pinMode(4,OUTPUT);  // res_pos2
  pinMode(5,OUTPUT);  // res_pos3
  pinMode(6,OUTPUT);  // valid0
  pinMode(7,OUTPUT);  // valid1
  pinMode(8,OUTPUT);  // valid2
  pinMode(9,OUTPUT);  // valid3
  digitalWrite(A2,LOW);  // flip>0
  digitalWrite(A3,LOW);  // total_bits>0
  digitalWrite(A4,LOW);  // case2
  digitalWrite(2,LOW);  // res_pos0
  digitalWrite(3,LOW);  // res_pos1
  digitalWrite(4,LOW);  // res_pos2
  digitalWrite(5,LOW);  // res_pos3
  digitalWrite(6,LOW);  // valid0 / UNKNOWN
  digitalWrite(7,LOW);  // valid1 / T0
  digitalWrite(8,LOW);  // valid2 / OK
  digitalWrite(9,LOW);  // valid3 / DONE


//  wait_SM("init end");

  // Vorbelegungen ohne cli() ok, weil erst
  //  anschließend die Interrupt-Routine scharf wird
    for (byte nn=0; nn<N_DATASETS; nn++)
      res_valid[nn]=0; // Flags der Datenbaenke

    for (byte nnn=0; nnn<N_DATASETS; nnn++)
      for (byte nn=0; nn<200; nn++)
        pulswidths[nnn][nn]=0;
      
    res_pos=0;
    orscV3.setdb(0);  // first buffer set
    digitalWrite(2+res_pos,HIGH);
    digitalWrite(2+1,LOW);
    digitalWrite(2+2,LOW);
    digitalWrite(2+3,LOW);
  
  // Empfaenger-Interrupt mit CMP1 fuer Teensy aktivieren:
    CMP1_Init();
    Comp1SetOn(0, 7);
    Comp1EnableIntrpt(CHANGE);

    Serial.println("es ist ein Teensy");
    intrPin=23;

#ifdef WGR800 // defined in OregonDecodeV3_2.h
  Serial.println("Pulssequenzen wie WGR800");
#else
  Serial.println("Pulssequenzen wie TFA-Sensor");
#endif

  Serial.println("  init done");

  tm_wd=millis();
  tm=tm_wd;
  talkInterv=millis();
}

void loop() { // ---------------------------------------------------------
  // to read a variable which the interrupt code writes, we
  // must temporarily disable interrupts, to be sure it will
  // not change while we are reading.  To minimize the time
  // with interrupts off, just quickly make a copy, and then
  // use the copy while allowing the interrupt to keep working
  boolean dvalid;
  uint16_t dtmin_c, dtmax_c;
  
  // --------------------------------------
    delay(1000);
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
//    Comp1DisableIntrpt();
    {
      respos_cpy = res_pos;
    }
//    Comp1EnableIntrpt(CHANGE);
      for (n=0; n<N_DATASETS; n++) {
//        digitalWrite(6+n,HIGH);
        delay(4);
        ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
//        Comp1DisableIntrpt();
        {
          dvalid = (res_valid[n] == 1); 
        }
//        Comp1EnableIntrpt(CHANGE);
        if (dvalid)  {  // data set is complete; therefore intr does not touch it
          
          ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
//          Comp1DisableIntrpt();
          {
            dbcnt_res = dbcount[n];
            dtmin_c = dtmin;
            dtmax_c = dtmax;
            dtmin=255;
            dtmax=0;
          }
//          Comp1EnableIntrpt(CHANGE);
          ts=micros();
            Serial.print(F("Messzeit: "));Serial.print(Scount); // irgendwas, Scount
                  Scount=0;
            Serial.print(F("  unk/tot: ")); Serial.print(cnt_unk);
            Serial.print("/");Serial.println(cnt_tot);
//            idxp=idx[n];
//            for (byte ii=0; ii<200; ii++) {   
//              Serial.print(pulswidths[n][idxp++]);Serial.print(" ");
//              if (idxp>199) idxp=0;
//              if (ii%20 == 19) Serial.println();
//            }
//            Serial.println();
//            Serial.flush();

            Serial.print("print ");Serial.print(micros()-ts); Serial.print("us |");
            Serial.print(millis()-tm_wd);Serial.print(": ");
            
            dtyp = readRFData(n, &v16, &v8, &lowbatt, false, false);
            
            switch (dtyp) {
              case 0:  // Temperaturdaten
                tm_wd=millis();
                snprintf(outstr,25,"n=%d dbcnt=%d T%04dH%02dB%d",n,dbcnt_res,v16,v8,lowbatt);
                Serial.print(outstr);
                mark_dataReceived();
                cnt_tot++;
                break;
              case 1: // Winddaten WGR800
                tm_wd=millis();
                snprintf(outstr,25,"n=%d dbcnt=%d W%03dR%02dB%d",n,dbcnt_res,v16,v8,lowbatt);
                Serial.print(outstr);
                mark_dataReceived();
                cnt_tot++;
                break;
              case 2:  // unbekanntes Signal
                digitalWrite(A4,HIGH);
                tm_wd=millis();
                Serial.print(F(" n="));Serial.print(n);
                Serial.print(F(" dbcnt="));Serial.print(dbcnt_res);
                Serial.print(F(" unbekanntes Signal "));
                tone(A0,2000,300);

                readRFData(n, &v16, &v8, &lowbatt, false, true);  // prints data set in hex
                delay(2);
                cnt_unk++;
                digitalWrite(A4,LOW);
            } // switch
            Serial.println();
            // on T3.2 an interrupt takes max. 5 usec
            Serial.print(F(" Interrupt-Dauer von "));
            Serial.print(dtmin_c);
            Serial.print(" bis ");Serial.println(dtmax_c);
            Serial.println();
  
            res_valid[n] = 0;   // wenn es 1 war, dann wurde in der Interrupt-Routine
//                                // bestimmt nicht daran gearbeitet
            
          } // if (res_valid..)
          else delay(2);  // damit leere Loop am Stateanalyser sichtbar wird       
//          tm=millis();
//        digitalWrite(6+n,LOW);
      } // for n
    Scount++;
}
